/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2012 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Chung Leong <cleong@cal.berkeley.edu>                        |
  +----------------------------------------------------------------------+
*/

/* $Id$ */

#include "qb.h"
#include "qb_interpreter_msvc.h"


void qb_main(qb_interpreter_context *__restrict cxt) {
	register void *__restrict handler;
	register int8_t *__restrict ip;
	uint32_t windows_timeout_check_counter = 0;
	volatile zend_bool *windows_timed_out_pointer = cxt->windows_timed_out_pointer;
	
	{
		handler = ((qb_instruction *) cxt->instruction_pointer)->next_handler;
		ip = cxt->instruction_pointer + sizeof(qb_instruction);
	}
	
	do {
		switch((intptr_t) handler) {
			case QB_NOP:
#define INSTR		((qb_instruction * __restrict) ip)
			{
				handler = INSTR->next_handler;
				ip += sizeof(qb_instruction);
				break;
			}
#undef INSTR
			
			case QB_JMP:
#define INSTR		((qb_instruction_jump * __restrict) ip)
			{
				handler = INSTR->next_handler;
				ip = INSTR->instruction_pointer;
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
			
			case QB_RET:
			{
				cxt->exit_type = QB_VM_RETURN;
				return;
			}
			
			case QB_INTR:
#define INSTR		((qb_instruction * __restrict) ip)
			{
				handler = INSTR->next_handler;
				cxt->instruction_pointer = ip;
				cxt->exit_type = QB_VM_YIELD;
				return;
				ip += sizeof(qb_instruction);
				break;
			}
#undef INSTR
			
			case QB_EXIT_I32:
#define INSTR		((qb_instruction_exit_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
			{
				cxt->exit_status_code = op1;
				cxt->exit_type = QB_VM_FORK;
				return;
			}
#undef INSTR
#undef op1
			
			case QB_FCALL_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_ARR_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					qb_dispatch_function_call(cxt, op1, op2_ptr, op2_count, op3, line_id);
					if(UNEXPECTED(cxt->exception_encountered)) {
						cxt->exit_type = QB_VM_EXCEPTION;
						return;
					}
				}
				ip += sizeof(qb_instruction_SCA_ARR_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2_ptr
#undef op2_count
#undef op3
			
			case QB_END_STATIC:
#define INSTR		((qb_instruction * __restrict) ip)
			{
				handler = INSTR->next_handler;
				cxt->function->instruction_start = ip;
				ip += sizeof(qb_instruction);
				break;
			}
#undef INSTR
			
			case QB_FORK_U32:
#define INSTR		((qb_instruction_exit_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
			{
				cxt->instruction_pointer = ip + sizeof(void *);
				cxt->fork_count = op1;
				cxt->exit_type = QB_VM_FORK;
				return;
			}
#undef INSTR
#undef op1
			
			case QB_FORK_ID_U32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((uint32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_fork_identifier_U32(cxt, &res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_FORK_ID_U32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_fork_identifier_U32(cxt, &res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_FORK_CNT_U32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((uint32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_fork_count_U32(cxt, &res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_FORK_CNT_U32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_fork_count_U32(cxt, &res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_RESUME:
#define INSTR		((qb_instruction * __restrict) ip)
			{
				handler = INSTR->next_handler;
				ip += sizeof(qb_instruction);
				break;
			}
#undef INSTR
			
			case QB_SPOON:
#define INSTR		((qb_instruction * __restrict) ip)
			{
				handler = INSTR->next_handler;
				cxt->instruction_pointer = ip;
				cxt->exit_type = QB_VM_SPOON;
				return;
				ip += sizeof(qb_instruction);
				break;
			}
#undef INSTR
			
			case QB_INC_IF_GT_U32_U32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_increment_branch_on_greater_than_U32(&condition, op1, &res);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_IF_LT_U32_U32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_U32_U32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U32_U32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U32_U32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IDX_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_index_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_IDX_GUARD_ADD_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_index_add_U32(cxt, op1, op2, op3, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_IDX_GUARD_MUL_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_index_multiply_U32(cxt, op1, op2, op3, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_IDX_GUARD_MAC_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand5.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_index_multiply_add_U32(cxt, op1, op2, op3, op4, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef op4
#undef res
			
			case QB_SZ_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_size_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_SZ_GUARD_EX_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_size_exact_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_EXT_GUARD_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_extent_U32(cxt, op1, op2, op3, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
			
			case QB_EXT_GUARD_MUL_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand5.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_extent_multiply_U32(cxt, op1, op2, op3, op4, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef op4
#undef res
			
			case QB_EXT_GUARD_SUB_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_extent_subtract_U32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_EXT_GUARD_SUB_MUL_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_guard_extent_subtract_multiply_U32(cxt, op1, op2, op3, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_IDX_CHECK_ADD_U32_U32_U32_I32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((int32_t *) INSTR->operand4.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand5.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_check_index_add_U32(cxt, op1, op2, op3, &op4, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef op4
#undef res
			
			case QB_IDX_CHECK_MUL_U32_U32_U32_I32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((int32_t *) INSTR->operand4.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand5.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_check_index_multiply_U32(cxt, op1, op2, op3, &op4, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef op4
#undef res
			
			case QB_IDX_CHECK_MAC_U32_U32_U32_U32_I32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((int32_t *) INSTR->operand5.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand6.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_check_index_multiply_add_U32(cxt, op1, op2, op3, op4, &op5, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef res
			
			case QB_IDX_ACCOM_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define op4	INSTR->operand4
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_index_U32(cxt, op1, &op2, op3, op4);
				ip += sizeof(qb_instruction_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
			
			case QB_IDX_ACCOM_MUL_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_U32_U32_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	INSTR->operand5
#define op6	INSTR->operand6
#define res	((uint32_t *) INSTR->operand7.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_index_multiply_U32(cxt, op1, &op2, op3, &op4, op5, op6, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_U32_U32_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res
			
			case QB_IDX_ACCOM_PUSH_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_U32_U32_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define op3	INSTR->operand3
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_push_U32(cxt, &op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_U32_U32_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_ACCOM_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define op4	INSTR->operand4
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_U32(cxt, op1, &op2, op3, op4);
				ip += sizeof(qb_instruction_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
			
			case QB_SZ_ACCOM_DIM_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	INSTR->operand5
#define op6	INSTR->operand6
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_update_dimension_U32(cxt, op1, &op2, &op3, op4, op5, op6);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
			
			case QB_SZ_ACCOM_DIM2_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	INSTR->operand9
#define op10	INSTR->operand10
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_copy_dimension_2_U32(cxt, op1, op2, &op3, &op4, op5, op6, &op7, &op8, op9, op10);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
			
			case QB_SZ_ACCOM_DIM3_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	INSTR->operand13
#define op14	INSTR->operand14
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_copy_dimension_3_U32(cxt, op1, op2, &op3, &op4, op5, op6, &op7, &op8, op9, op10, &op11, &op12, op13, op14);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
			
			case QB_SZ_ACCOM_DIM4_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	INSTR->operand17
#define op18	INSTR->operand18
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_copy_dimension_4_U32(cxt, op1, op2, &op3, &op4, op5, op6, &op7, &op8, op9, op10, &op11, &op12, op13, op14, &op15, &op16, op17, op18);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
			
			case QB_SZ_ACCOM_DIM5_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	INSTR->operand21
#define op22	INSTR->operand22
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_copy_dimension_5_U32(cxt, op1, op2, &op3, &op4, op5, op6, &op7, &op8, op9, op10, &op11, &op12, op13, op14, &op15, &op16, op17, op18, &op19, &op20, op21, op22);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
			
			case QB_SZ_ACCOM_DIM6_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	INSTR->operand25
#define op26	INSTR->operand26
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_copy_dimension_6_U32(cxt, op1, op2, &op3, &op4, op5, op6, &op7, &op8, op9, op10, &op11, &op12, op13, op14, &op15, &op16, op17, op18, &op19, &op20, op21, op22, &op23, &op24, op25, op26);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef op26
			
			case QB_SZ_ACCOM_DIM7_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	((uint32_t *) INSTR->operand25.data_pointer)[0]
#define op26	((uint32_t *) INSTR->operand26.data_pointer)[0]
#define op27	((uint32_t *) INSTR->operand27.data_pointer)[0]
#define op28	((uint32_t *) INSTR->operand28.data_pointer)[0]
#define op29	INSTR->operand29
#define op30	INSTR->operand30
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_copy_dimension_7_U32(cxt, op1, op2, &op3, &op4, op5, op6, &op7, &op8, op9, op10, &op11, &op12, op13, op14, &op15, &op16, op17, op18, &op19, &op20, op21, op22, &op23, &op24, op25, op26, &op27, &op28, op29, op30);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef op26
#undef op27
#undef op28
#undef op29
#undef op30
			
			case QB_SZ_ACCOM_DIM8_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32 * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	((uint32_t *) INSTR->operand25.data_pointer)[0]
#define op26	((uint32_t *) INSTR->operand26.data_pointer)[0]
#define op27	((uint32_t *) INSTR->operand27.data_pointer)[0]
#define op28	((uint32_t *) INSTR->operand28.data_pointer)[0]
#define op29	((uint32_t *) INSTR->operand29.data_pointer)[0]
#define op30	((uint32_t *) INSTR->operand30.data_pointer)[0]
#define op31	((uint32_t *) INSTR->operand31.data_pointer)[0]
#define op32	((uint32_t *) INSTR->operand32.data_pointer)[0]
#define op33	INSTR->operand33
#define op34	INSTR->operand34
			{
				handler = INSTR->next_handler;
				qb_do_accommodate_size_copy_dimension_8_U32(cxt, op1, op2, &op3, &op4, op5, op6, &op7, &op8, op9, op10, &op11, &op12, op13, op14, &op15, &op16, op17, op18, &op19, &op20, op21, op22, &op23, &op24, op25, op26, &op27, &op28, op29, op30, &op31, &op32, op33, op34);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_U32);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef op26
#undef op27
#undef op28
#undef op29
#undef op30
#undef op31
#undef op32
#undef op33
#undef op34
			
			case QB_SZ_CHOOSE_2_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand5.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_choose_size_larger_of_two_U32(op1, op2, op3, op4, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res
			
			case QB_SZ_CHOOSE_2_TOP_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_choose_size_larger_of_two_top_level_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SZ_CHOOSE_3_U32_U32_U32_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand7.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_choose_size_largest_of_three_U32(op1, op2, op3, op4, op5, op6, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res
			
			case QB_SZ_CHOOSE_3_TOP_U32_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_choose_size_largest_of_three_top_level_U32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MUL_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_U32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_U32_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_U32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_U32_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_U32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_U32_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_multiple_times_U32(op1_ptr, op1_count, op2_ptr, op2_count, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_U32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_U32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_U32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_U32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_U32_U32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U32_U32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U32_U32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_U32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_U32_U32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U32_U32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U32_U32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_U32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_SET_U32_U32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_U32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_U32_U32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_U32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_U32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_U32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_U32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_U32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_U32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U32_U32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_U32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_U32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_U32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_U32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_U32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_U32_U32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_U32_U32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U32_U32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U32_U32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U32_U32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U32_U32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U32_U32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U32_U32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_U32_U32_S32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((uint32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_U32_U32_S32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((uint32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_U32_U32_S32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_U32_U32_S32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_APAD_U32_S32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_pad_count_U32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ARAND_U32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_random_U32(cxt, op1, op2, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_ARAND_U32_U32_U32_ELE:
#define INSTR		((qb_instruction_SCA_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res_ptr		(((uint32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_random_U32(cxt, op1, op2, res_ptr);
				ip += sizeof(qb_instruction_SCA_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_IF_T_I32_SCA:
#define INSTR		((qb_instruction_branch_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = op1;
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
			
			case QB_IF_T_I32_ELE:
#define INSTR		((qb_instruction_branch_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = op1;
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
			
			case QB_IF_F_I32_SCA:
#define INSTR		((qb_instruction_branch_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = !op1;
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
			
			case QB_IF_F_I32_ELE:
#define INSTR		((qb_instruction_branch_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = !op1;
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
			
			case QB_IF_EQ_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_EQ_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S32_S32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S32_S32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S32_S32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S32_S32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_9_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_9_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_17_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_17_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_33_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_33_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_65_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_65_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_129_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_129_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_257_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_257_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_513_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_513_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I32_I32_SCA:
#define INSTR		((qb_instruction_branch_table_1025_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I32_I32_ELE:
#define INSTR		((qb_instruction_branch_table_1025_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_INC_I32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I32(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I32(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_increment_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_I32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I32(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I32(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_decrement_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_CBOOL_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_CBOOL_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_ELE_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOV_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_multiple_times_I32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_add_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_S32_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_S32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_S32_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((int32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_S32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_S32_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_multiple_times_S32(op1_ptr, op1_count, op2_ptr, op2_count, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_S32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S32(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_S32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BW_AND_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_and_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_OR_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_or_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_XOR_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_xor_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_NOT_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_not_multiple_times_I32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_EQ_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_multiple_times_I32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_NE_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_multiple_times_I32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_S32_S32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S32_S32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S32_S32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_S32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_S32_S32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S32_S32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S32_S32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_S32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_EQ_SET_I32_I32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_equal_I32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NE_SET_I32_I32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_not_equal_I32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_SET_S32_S32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_S32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_S32_S32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_S32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NOT_SET_I32_I32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_not_I32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ANY_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_any_I32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ANY_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_set_any_I32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ALL_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_all_I32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ALL_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_set_all_I32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AND_I32_I32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_logical_and_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_OR_I32_I32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_logical_or_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_XOR_I32_I32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_logical_xor_I32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NOT_I32_I32:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_logical_not_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_abs_multiple_times_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_S32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_S32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_S32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_S32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_S32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_S32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S32(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S32_S32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_S32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_S32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_S32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_S32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_S32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_S32_S32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_S32_S32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S32_S32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S32_S32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S32_S32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S32_S32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S32_S32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S32_S32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_S32_S32_S32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_S32_S32_S32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_S32_S32_S32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_S32_S32_S32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_AFIND_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I32_I32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_multiple_times_I32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_AFILL_U32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I32(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I32_I32_ELE:
#define INSTR		((qb_instruction_SCA_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I32(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I32_I32_MIO:
#define INSTR		((qb_instruction_SCA_ARR_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_multiple_times_I32(op1, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_APOS_I32_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_APOS_I32_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I32_I32_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I32_I32_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AREV_I32_U32_I32:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_reverse_I32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AUNIQ_I32_U32_I32:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_I32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_SZ_AUNIQ_I32_U32_U32:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_count_I32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_ACOL_I32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_array_column_I32(cxt, op1_ptr, op1_count, op2, op3, op4, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ADIFF_I32_I32_U32_I32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_ADIFF_I32_I32_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_count_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AISECT_I32_I32_U32_I32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_AISECT_I32_I32_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_count_I32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_SHUFFLE_U32_I32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_shuffle_I32(cxt, op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I32(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I32(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SZ_ASLICE_S32_S32_U32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_count_I32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_ASLICE_S32_S32_U32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_count_I32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_AREPLACE_I32_S32_S32_U32_U32_U32_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int32_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I32(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_AREPLACE_I32_S32_S32_U32_U32_U32_I32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int32_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I32(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_ARESIZE1_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_1_I32(cxt, op1, &op2, &op3, op4, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE2_U32_U32_U32_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	INSTR->operand7
#define res_ptr		(((int32_t *) INSTR->operand8.data_pointer) + INSTR->operand8.index_pointer[0])
#define res_count	INSTR->operand8.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_2_I32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE3_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	INSTR->operand10
#define res_ptr		(((int32_t *) INSTR->operand11.data_pointer) + INSTR->operand11.index_pointer[0])
#define res_count	INSTR->operand11.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_3_I32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE4_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	INSTR->operand13
#define res_ptr		(((int32_t *) INSTR->operand14.data_pointer) + INSTR->operand14.index_pointer[0])
#define res_count	INSTR->operand14.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_4_I32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE5_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	INSTR->operand16
#define res_ptr		(((int32_t *) INSTR->operand17.data_pointer) + INSTR->operand17.index_pointer[0])
#define res_count	INSTR->operand17.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_5_I32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE6_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	INSTR->operand19
#define res_ptr		(((int32_t *) INSTR->operand20.data_pointer) + INSTR->operand20.index_pointer[0])
#define res_count	INSTR->operand20.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_6_I32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE7_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	INSTR->operand22
#define res_ptr		(((int32_t *) INSTR->operand23.data_pointer) + INSTR->operand23.index_pointer[0])
#define res_count	INSTR->operand23.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_7_I32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE8_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	INSTR->operand25
#define res_ptr		(((int32_t *) INSTR->operand26.data_pointer) + INSTR->operand26.index_pointer[0])
#define res_count	INSTR->operand26.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_8_I32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, &op23, &op24, op25, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef res_ptr
#undef res_count
			
			case QB_APAD_I32_S32_I32_I32:
#define INSTR		((qb_instruction_ARR_SCA_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_pad_I32(op1_ptr, op1_count, op2, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_IF_EQ_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_EQ_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_F32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_F32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_F32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_F32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_F32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_F32(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_9_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_9_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_17_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_17_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_33_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_33_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_65_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_65_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_129_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_129_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_257_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_257_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_513_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_513_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_F32_F32_SCA:
#define INSTR		((qb_instruction_branch_table_1025_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_F32_F32_ELE:
#define INSTR		((qb_instruction_branch_table_1025_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_INC_F32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_F32(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_F32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_increment_F32(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_increment_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_F32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_F32(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_F32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_F32(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_decrement_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_CBOOL_F32_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_CBOOL_F32_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_ELE_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOV_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_copy_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_multiple_times_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_add_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_F32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_F32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_DIV_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_divide_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_DIV_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_FLR_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_floored_division_modulo_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_FLR_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_floored_division_modulo_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_FLR_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_floored_division_modulo_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_negate_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EQ_F32_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_F32_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_F32_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_NE_F32_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_F32_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_F32_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_F32_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_F32_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_F32_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_F32_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_F32_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_F32_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_EQ_SET_F32_F32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_equal_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NE_SET_F32_F32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_not_equal_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_SET_F32_F32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_F32_F32_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ABS_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_abs_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_abs_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_abs_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SIN_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sin_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIN_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sin_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIN_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sin_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ASIN_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_asin_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASIN_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_asin_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASIN_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_asin_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_COS_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cos_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COS_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cos_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COS_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cos_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ACOS_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_acos_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOS_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_acos_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOS_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_acos_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_TAN_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_tan_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TAN_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_tan_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TAN_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_tan_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ATAN_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_atan_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATAN_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_atan_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATAN_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_atan_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ATAN2_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_atan2_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ATAN2_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_atan2_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ATAN2_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_atan2_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SINH_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sinh_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SINH_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sinh_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SINH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sinh_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ASINH_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_asinh_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASINH_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_asinh_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASINH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_asinh_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_COSH_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cosh_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COSH_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cosh_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COSH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cosh_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ACOSH_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_acosh_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOSH_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_acosh_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOSH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_acosh_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_TANH_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_tanh_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TANH_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_tanh_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TANH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_tanh_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ATANH_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_atanh_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATANH_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_atanh_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATANH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_atanh_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CEIL_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_ceil_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_CEIL_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_ceil_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_CEIL_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_ceil_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_FLOOR_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_floor_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FLOOR_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_floor_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FLOOR_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_floor_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RINT_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_rint_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RINT_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_rint_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RINT_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rint_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ROUND_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_round_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ROUND_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_round_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ROUND_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_round_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG1P_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log1p_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG1P_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log1p_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG1P_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log1p_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG2_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log2_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG2_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log2_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG2_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log2_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG10_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log10_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG10_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log10_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG10_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log10_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EXP_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_exp_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_exp_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_exp_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EXPM1_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_exp_m1_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXPM1_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_exp_m1_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXPM1_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_exp_m1_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EXP2_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_exp2_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP2_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_exp2_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP2_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_exp2_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_POW_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_pow_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_POW_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_pow_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_POW_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_pow_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SQRT_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sqrt_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SQRT_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sqrt_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SQRT_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sqrt_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HYPOT_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_hypot_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_HYPOT_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_hypot_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_HYPOT_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hypot_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LCG_F32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_lcg_F32(cxt, &res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_LCG_F32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_lcg_F32(cxt, &res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_LCG_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_lcg_multiple_times_F32(cxt, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_FIN_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_is_finite_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FIN_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_is_finite_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FIN_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_is_finite_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INF_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_is_infinite_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_INF_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_is_infinite_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_INF_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_is_infinite_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NAN_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_is_na_n_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NAN_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_is_na_n_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NAN_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_is_na_n_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RSQRT_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_rsqrt_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RSQRT_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_rsqrt_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RSQRT_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rsqrt_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CLAMP_F32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clamp_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_CLAMP_F32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_clamp_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_CLAMP_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_clamp_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_FRACT_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_fract_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FRACT_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_fract_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FRACT_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_fract_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIX_F32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_mix_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MIX_F32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_mix_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MIX_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_mix_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SIGN_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sign_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIGN_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sign_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIGN_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sign_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_STEP_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_step_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_STEP_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_step_F32(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_STEP_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_step_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SSTEP_F32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_smooth_step_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SSTEP_F32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_smooth_step_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SSTEP_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_smooth_step_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_RAD2DEG_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_radian_to_degree_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RAD2DEG_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_radian_to_degree_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RAD2DEG_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_radian_to_degree_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DEG2RAD_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_degree_to_radian_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_DEG2RAD_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_degree_to_radian_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_DEG2RAD_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_degree_to_radian_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ROUND_F32_I32_I32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_round_to_precision_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_ROUND_F32_I32_I32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_round_to_precision_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_ROUND_F32_I32_I32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_round_to_precision_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_F32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_F32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_F32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_F32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_F32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_F32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_F32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_F32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_F32_F32_F32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_F32_F32_F32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_F32(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_AFIND_F32_F32_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_F32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_F32_F32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_F32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_F32_F32_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_AFILL_U32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_F32(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_F32_F32_ELE:
#define INSTR		((qb_instruction_SCA_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_F32(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_F32_F32_MIO:
#define INSTR		((qb_instruction_SCA_ARR_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_multiple_times_F32(op1, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_APOS_F32_F32_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_APOS_F32_F32_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_F32_F32_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_F32_F32_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AREV_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_reverse_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AUNIQ_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_SZ_AUNIQ_F32_U32_U32:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_count_F32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_ACOL_F32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_array_column_F32(cxt, op1_ptr, op1_count, op2, op3, op4, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ADIFF_F32_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_ADIFF_F32_F32_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_count_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AISECT_F32_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_AISECT_F32_F32_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_count_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_SHUFFLE_U32_F32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_shuffle_F32(cxt, op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_F32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_F32(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_F32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_F32(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_AREPLACE_F32_S32_S32_U32_U32_U32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((float32_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_F32(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_AREPLACE_F32_S32_S32_U32_U32_U32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((float32_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_F32(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_ARESIZE1_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_1_F32(cxt, op1, &op2, &op3, op4, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE2_U32_U32_U32_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	INSTR->operand7
#define res_ptr		(((float32_t *) INSTR->operand8.data_pointer) + INSTR->operand8.index_pointer[0])
#define res_count	INSTR->operand8.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_2_F32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE3_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	INSTR->operand10
#define res_ptr		(((float32_t *) INSTR->operand11.data_pointer) + INSTR->operand11.index_pointer[0])
#define res_count	INSTR->operand11.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_3_F32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE4_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	INSTR->operand13
#define res_ptr		(((float32_t *) INSTR->operand14.data_pointer) + INSTR->operand14.index_pointer[0])
#define res_count	INSTR->operand14.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_4_F32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE5_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	INSTR->operand16
#define res_ptr		(((float32_t *) INSTR->operand17.data_pointer) + INSTR->operand17.index_pointer[0])
#define res_count	INSTR->operand17.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_5_F32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE6_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	INSTR->operand19
#define res_ptr		(((float32_t *) INSTR->operand20.data_pointer) + INSTR->operand20.index_pointer[0])
#define res_count	INSTR->operand20.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_6_F32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE7_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	INSTR->operand22
#define res_ptr		(((float32_t *) INSTR->operand23.data_pointer) + INSTR->operand23.index_pointer[0])
#define res_count	INSTR->operand23.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_7_F32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE8_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	INSTR->operand25
#define res_ptr		(((float32_t *) INSTR->operand26.data_pointer) + INSTR->operand26.index_pointer[0])
#define res_count	INSTR->operand26.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_8_F32(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, &op23, &op24, op25, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef res_ptr
#undef res_count
			
			case QB_APAD_F32_S32_F32_F32:
#define INSTR		((qb_instruction_ARR_SCA_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_pad_F32(op1_ptr, op1_count, op2, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_4X_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_4x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_4X_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_4x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_4X_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_3X_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_3x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_3X_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_3x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_3X_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_2X_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_2x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_2X_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_2x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_2X_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_4X_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_4x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_4X_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_4x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_4X_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_3X_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_3x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_3X_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_3x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_3X_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_2X_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_2x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_2X_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_2x_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_2X_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_F32_U32_U32_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_F32_U32_U32_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float32_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_F32(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_F32_U32_U32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_alpha_blend_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_alpha_blend_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_alpha_blend_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_alpha_blend_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_PREMULT_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_apply_premultiplication_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_PREMULT_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_apply_premultiplication_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_UNPREMULT_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_remove_premultiplication_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_UNPREMULT_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_remove_premultiplication_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsv_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsv_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsv_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsv_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsv2rgb_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsv2rgb_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsv2rgb_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsv2rgb_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsl_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsl_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsl_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsl_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsl2rgb_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsl2rgb_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsl2rgb_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsl2rgb_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_4x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_matrix_4x_multiple_times_column_major_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_4x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_4x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_4x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_4x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transpose_matrix_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_invert_matrix_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MDET_4X_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_4x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_4X_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_4x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_determinant_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DOT_4X_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_4x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_4X_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_4x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_dot_product_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LEN_4X_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_4x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_4X_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_4x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_length_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DIS_4X_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_4x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_4X_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_4x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_distance_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_normalize_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_4X_F32_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cross_product_4x_F32(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_4X_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cross_product_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_face_forward_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_reflect_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFR_4X_F32_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_4x_F32(op1_ptr, op2_ptr, op3, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_4X_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_refract_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_4x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_4x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_4x_row_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_4x_multiple_times_row_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_4x_multiple_times_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_4X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_4x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_4X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_4X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_4x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_4X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INC_4X_F32:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_4x_F32(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_INC_4X_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_4x_multiple_times_F32(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_4X_F32:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_4x_F32(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_4X_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_4x_multiple_times_F32(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_MAC_4X_F32_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_4x_F32(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_4X_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_4x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_3x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_matrix_3x_multiple_times_column_major_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_3x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_3x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_3x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_3x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_PAD_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_padded_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_PAD_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_padded_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_PAD_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_padded_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transpose_matrix_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_invert_matrix_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MDET_3X_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_3x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_3X_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_3x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_determinant_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DOT_3X_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_3x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_3X_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_3x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_dot_product_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LEN_3X_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_3x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_3X_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_3x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_length_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DIS_3X_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_3x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_3X_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_3x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_distance_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_normalize_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cross_product_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cross_product_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_face_forward_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_reflect_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFR_3X_F32_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_3x_F32(op1_ptr, op2_ptr, op3, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_3X_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_refract_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_3x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_3x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_3x_row_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_3x_multiple_times_row_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_3x_multiple_times_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_3X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_3x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_3X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_3X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_3x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_3X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INC_3X_F32:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_3x_F32(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_INC_3X_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_3x_multiple_times_F32(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_3X_F32:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_3x_F32(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_3X_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_3x_multiple_times_F32(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_MAC_3X_F32_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_3x_F32(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_3X_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_3x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_2x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_matrix_2x_multiple_times_column_major_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_2x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_2x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_2x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_2x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_2X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_2X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transpose_matrix_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_2X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_2X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_invert_matrix_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MDET_2X_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_2x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_2X_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_2x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_2X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_2x_multiple_times_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DOT_2X_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_2x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_2X_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_2x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_dot_product_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LEN_2X_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_2x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_2X_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_2x_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_2X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_length_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DIS_2X_F32_F32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_2x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_2X_F32_F32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_2x_F32(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_distance_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_2X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_2X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_normalize_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cross_product_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cross_product_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_face_forward_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_reflect_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFR_2X_F32_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_2x_F32(op1_ptr, op2_ptr, op3, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_2X_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_refract_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_2x_column_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_2x_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_2x_row_major_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_2x_multiple_times_row_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_2X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_2X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_2x_multiple_times_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_2X_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_2X_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_2X_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_2X_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INC_2X_F32:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_2x_F32(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_INC_2X_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_increment_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_2X_F32:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_2x_F32(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_2X_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_2x_multiple_times_F32(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_MAC_2X_F32_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_2x_F32(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_2X_F32_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_F32_F32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, op5, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_F32_F32_U32_U32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, op5, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_F32_F32_U32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_F32_F32_U32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_F32_F32_U32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_F32_F32_U32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_multiple_times_column_major_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_F32_U32_U32_F32:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_F32(op1_ptr, op1_count, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_F32_U32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_multiple_times_F32(op1_ptr, op1_count, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_MINV_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MINV_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_multiple_times_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MDET_F32_U32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_F32(op1_ptr, op1_count, op2, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MDET_F32_U32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_F32(op1_ptr, op1_count, op2, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MDET_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_multiple_times_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_DOT_F32_F32_U32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DOT_F32_F32_U32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DOT_F32_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_LEN_F32_U32_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_F32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_LEN_F32_U32_F32_ELE:
#define INSTR		((qb_instruction_ARR_SCA_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_F32(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_LEN_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_multiple_times_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_DIS_F32_F32_U32_F32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DIS_F32_F32_U32_F32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ELE * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float32_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DIS_F32_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_NORM_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_NORM_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_multiple_times_F32(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_FORE_F32_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_FORE_F32_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFL_F32_F32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFL_F32_F32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_F32_F32_U32_U32_F32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_REFR_F32_F32_U32_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_multiple_times_F32(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_CABS_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_abs_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CABS_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_abs_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CARG_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_argument_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CARG_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_argument_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CMUL_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_multiply_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CMUL_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_multiply_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CDIV_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_divide_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CDIV_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_divide_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CEXP_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_exp_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CEXP_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_exp_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CLOG_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_log_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CLOG_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_log_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSQRT_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_square_root_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSQRT_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_square_root_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CPOW_F32_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_pow_2x_F32(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CPOW_F32_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_pow_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CSIN_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_sin_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSIN_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_sin_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOS_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_cos_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOS_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_cos_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTAN_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_tan_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTAN_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_tan_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSINH_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_sinh_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSINH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_sinh_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOSH_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_cosh_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOSH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_cosh_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTANH_F32_F32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_tanh_2x_F32(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTANH_F32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_tanh_2x_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_IF_EQ_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_EQ_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_F64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_F64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_F64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_F64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_F64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_F64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_9_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_9_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_17_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_17_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_33_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_33_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_65_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_65_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_129_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_129_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_257_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_257_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_513_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_513_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_F64_F64_SCA:
#define INSTR		((qb_instruction_branch_table_1025_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_F64_F64_ELE:
#define INSTR		((qb_instruction_branch_table_1025_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_INC_F64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_F64(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_F64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_increment_F64(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_increment_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_F64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_F64(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_F64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_F64(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_decrement_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_CBOOL_F64_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_CBOOL_F64_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_ELE_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOV_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_copy_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_multiple_times_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_add_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_F64_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_F64_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_DIV_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_divide_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_DIV_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_FLR_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_floored_division_modulo_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_FLR_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_floored_division_modulo_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOD_FLR_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_floored_division_modulo_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_negate_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EQ_F64_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_F64_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_F64_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_NE_F64_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_F64_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_F64_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_F64_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_F64_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_F64_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_F64_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_F64_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_F64_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_EQ_SET_F64_F64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_equal_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NE_SET_F64_F64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_not_equal_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_SET_F64_F64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_F64_F64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ABS_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_abs_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_abs_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_abs_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SIN_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sin_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIN_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sin_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIN_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sin_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ASIN_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_asin_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASIN_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_asin_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASIN_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_asin_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_COS_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cos_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COS_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cos_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COS_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cos_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ACOS_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_acos_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOS_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_acos_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOS_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_acos_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_TAN_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_tan_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TAN_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_tan_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TAN_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_tan_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ATAN_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_atan_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATAN_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_atan_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATAN_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_atan_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ATAN2_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_atan2_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ATAN2_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_atan2_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ATAN2_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_atan2_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SINH_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sinh_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SINH_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sinh_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SINH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sinh_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ASINH_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_asinh_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASINH_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_asinh_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ASINH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_asinh_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_COSH_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cosh_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COSH_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cosh_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_COSH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cosh_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ACOSH_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_acosh_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOSH_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_acosh_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ACOSH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_acosh_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_TANH_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_tanh_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TANH_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_tanh_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_TANH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_tanh_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ATANH_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_atanh_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATANH_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_atanh_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ATANH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_atanh_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CEIL_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_ceil_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_CEIL_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_ceil_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_CEIL_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_ceil_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_FLOOR_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_floor_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FLOOR_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_floor_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FLOOR_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_floor_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RINT_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_rint_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RINT_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_rint_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RINT_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rint_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ROUND_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_round_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ROUND_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_round_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ROUND_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_round_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG1P_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log1p_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG1P_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log1p_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG1P_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log1p_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG2_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log2_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG2_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log2_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG2_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log2_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_LOG10_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_log10_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG10_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_log10_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_LOG10_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_log10_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EXP_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_exp_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_exp_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_exp_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EXPM1_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_exp_m1_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXPM1_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_exp_m1_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXPM1_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_exp_m1_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_EXP2_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_exp2_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP2_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_exp2_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_EXP2_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_exp2_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_POW_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_pow_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_POW_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_pow_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_POW_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_pow_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SQRT_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sqrt_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SQRT_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sqrt_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SQRT_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sqrt_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HYPOT_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_hypot_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_HYPOT_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_hypot_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_HYPOT_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hypot_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LCG_F64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_lcg_F64(cxt, &res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_LCG_F64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_lcg_F64(cxt, &res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_LCG_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_lcg_multiple_times_F64(cxt, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_FIN_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_is_finite_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FIN_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_is_finite_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FIN_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_is_finite_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INF_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_is_infinite_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_INF_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_is_infinite_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_INF_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_is_infinite_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NAN_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_is_na_n_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NAN_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_is_na_n_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NAN_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_is_na_n_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RSQRT_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_rsqrt_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RSQRT_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_rsqrt_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RSQRT_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rsqrt_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CLAMP_F64_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clamp_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_CLAMP_F64_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_clamp_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_CLAMP_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_clamp_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_FRACT_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_fract_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FRACT_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_fract_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_FRACT_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_fract_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIX_F64_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_mix_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MIX_F64_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_mix_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MIX_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_mix_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SIGN_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_sign_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIGN_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_sign_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_SIGN_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sign_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_STEP_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_step_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_STEP_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_step_F64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_STEP_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_step_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SSTEP_F64_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_smooth_step_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SSTEP_F64_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_smooth_step_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SSTEP_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_smooth_step_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_RAD2DEG_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_radian_to_degree_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RAD2DEG_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_radian_to_degree_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_RAD2DEG_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_radian_to_degree_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DEG2RAD_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_degree_to_radian_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_DEG2RAD_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_degree_to_radian_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_DEG2RAD_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_degree_to_radian_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ROUND_F64_I32_I32_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_round_to_precision_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_ROUND_F64_I32_I32_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_round_to_precision_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_ROUND_F64_I32_I32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_round_to_precision_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_F64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_F64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_F64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_F64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_F64_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_F64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_F64_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_F64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_F64_F64_F64_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_F64_F64_F64_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_F64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_AFIND_F64_F64_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_F64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_F64_F64_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_F64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_F64_F64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_AFILL_U32_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_F64(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_F64_F64_ELE:
#define INSTR		((qb_instruction_SCA_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_F64(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_F64_F64_MIO:
#define INSTR		((qb_instruction_SCA_ARR_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_multiple_times_F64(op1, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_APOS_F64_F64_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_APOS_F64_F64_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_F64_F64_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_F64_F64_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AREV_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_reverse_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AUNIQ_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_SZ_AUNIQ_F64_U32_U32:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_count_F64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_ACOL_F64_U32_U32_U32_F64:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_array_column_F64(cxt, op1_ptr, op1_count, op2, op3, op4, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ADIFF_F64_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_ADIFF_F64_F64_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_count_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AISECT_F64_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_AISECT_F64_F64_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_count_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_SHUFFLE_U32_F64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_shuffle_F64(cxt, op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_F64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_F64(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_F64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_F64(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_AREPLACE_F64_S32_S32_U32_U32_U32_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((float64_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_F64(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_AREPLACE_F64_S32_S32_U32_U32_U32_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((float64_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_F64(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_ARESIZE1_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_1_F64(cxt, op1, &op2, &op3, op4, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE2_U32_U32_U32_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	INSTR->operand7
#define res_ptr		(((float64_t *) INSTR->operand8.data_pointer) + INSTR->operand8.index_pointer[0])
#define res_count	INSTR->operand8.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_2_F64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE3_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	INSTR->operand10
#define res_ptr		(((float64_t *) INSTR->operand11.data_pointer) + INSTR->operand11.index_pointer[0])
#define res_count	INSTR->operand11.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_3_F64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE4_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	INSTR->operand13
#define res_ptr		(((float64_t *) INSTR->operand14.data_pointer) + INSTR->operand14.index_pointer[0])
#define res_count	INSTR->operand14.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_4_F64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE5_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	INSTR->operand16
#define res_ptr		(((float64_t *) INSTR->operand17.data_pointer) + INSTR->operand17.index_pointer[0])
#define res_count	INSTR->operand17.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_5_F64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE6_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	INSTR->operand19
#define res_ptr		(((float64_t *) INSTR->operand20.data_pointer) + INSTR->operand20.index_pointer[0])
#define res_count	INSTR->operand20.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_6_F64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE7_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	INSTR->operand22
#define res_ptr		(((float64_t *) INSTR->operand23.data_pointer) + INSTR->operand23.index_pointer[0])
#define res_count	INSTR->operand23.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_7_F64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE8_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	INSTR->operand25
#define res_ptr		(((float64_t *) INSTR->operand26.data_pointer) + INSTR->operand26.index_pointer[0])
#define res_count	INSTR->operand26.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_8_F64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, &op23, &op24, op25, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef res_ptr
#undef res_count
			
			case QB_APAD_F64_S32_F64_F64:
#define INSTR		((qb_instruction_ARR_SCA_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_pad_F64(op1_ptr, op1_count, op2, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_4X_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_4x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_4X_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_4x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_4X_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_3X_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_3x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_3X_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_3x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_3X_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_2X_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_2x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_2X_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_2x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_2X_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_nearest_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_NN_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_nearest_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_4X_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_4x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_4X_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_4x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_4X_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_3X_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_3x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_3X_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_3x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_3X_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_2X_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_2x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_2X_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_2x_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_2X_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_F64_U32_U32_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[0]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_F64_U32_U32_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
#define op5	((float64_t *) INSTR->operand5.data_pointer)[INSTR->operand5.index_pointer[0]]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sample_bilinear_F64(op1_ptr, op2, op3, op4, op5, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_SAMPLE_BL_F64_U32_U32_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define op4_count	INSTR->operand4.count_pointer[0]
#define op5_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_sample_bilinear_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4_ptr
#undef op4_count
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_alpha_blend_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_alpha_blend_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_alpha_blend_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BLEND_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_alpha_blend_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_PREMULT_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_apply_premultiplication_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_PREMULT_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_apply_premultiplication_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_UNPREMULT_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_remove_premultiplication_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_UNPREMULT_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_remove_premultiplication_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsv_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsv_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsv_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSV_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsv_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsv2rgb_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsv2rgb_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsv2rgb_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSV2RGB_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsv2rgb_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsl_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsl_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_rgb2hsl_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_RGB2HSL_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_rgb2hsl_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsl2rgb_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsl2rgb_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_hsl2rgb_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_HSL2RGB_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_hsl2rgb_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_4x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_matrix_4x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_4x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_vector_4x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_4x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_vector_by_matrix_4x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transpose_matrix_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_invert_matrix_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MDET_4X_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_4x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_4X_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_4x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_determinant_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DOT_4X_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_4x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_4X_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_4x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_4x_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LEN_4X_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_4x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_4X_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_4x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_length_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DIS_4X_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_4x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_4X_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_4x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_distance_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_normalize_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_4X_F64_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cross_product_4x_F64(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_4X_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cross_product_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_face_forward_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_reflect_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFR_4X_F64_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_4x_F64(op1_ptr, op2_ptr, op3, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_4X_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_refract_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_4x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transform_vector_4x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_4x_row_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transform_vector_4x_multiple_times_row_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_4x_multiple_times_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_4X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_4x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_4X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_4X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_4x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_4X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INC_4X_F64:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_4x_F64(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_INC_4X_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_4x_multiple_times_F64(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_4X_F64:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_4x_F64(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_4X_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_4x_multiple_times_F64(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_MAC_4X_F64_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_4x_F64(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_4X_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_4x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_3x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_matrix_3x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_3x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_vector_3x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_3x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_vector_by_matrix_3x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transpose_matrix_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_invert_matrix_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MDET_3X_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_3x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_3X_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_3x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_determinant_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DOT_3X_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_3x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_3X_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_3x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_3x_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LEN_3X_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_3x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_3X_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_3x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_length_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DIS_3X_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_3x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_3X_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_3x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_distance_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_normalize_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cross_product_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_cross_product_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_face_forward_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_reflect_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFR_3X_F64_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_3x_F64(op1_ptr, op2_ptr, op3, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_3X_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_refract_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_3x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transform_vector_3x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_3x_row_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transform_vector_3x_multiple_times_row_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_3x_multiple_times_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_3X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_3x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_3X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_3X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_3x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_3X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INC_3X_F64:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_3x_F64(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_INC_3X_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_3x_multiple_times_F64(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_3X_F64:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_3x_F64(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_3X_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_3x_multiple_times_F64(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_MAC_3X_F64_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_3x_F64(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_3X_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_3x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_2x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_matrix_2x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_2x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_matrix_by_vector_2x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_2x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_vector_by_matrix_2x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_2X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_2X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transpose_matrix_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_2X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MINV_2X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_invert_matrix_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MDET_2X_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_2x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_2X_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_2x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MDET_2X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_2x_multiple_times_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DOT_2X_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_2x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_2X_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_2x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DOT_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_2x_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LEN_2X_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_2x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_2X_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_2x_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_LEN_2X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_length_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_DIS_2X_F64_F64_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_2x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_2X_F64_F64_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_2x_F64(op1_ptr, op2_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_DIS_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_distance_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_2X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NORM_2X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_normalize_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cross_product_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CROSS_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cross_product_2x_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_FORE_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_face_forward_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFL_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_reflect_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_REFR_2X_F64_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_2x_F64(op1_ptr, op2_ptr, op3, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_2X_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((float64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_refract_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_2x_column_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_CM_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transform_vector_2x_multiple_times_column_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transform_vector_2x_row_major_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_TRAN_RM_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_transform_vector_2x_multiple_times_row_major_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_2X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_2X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_2x_multiple_times_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_divide_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_divide_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_2X_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_modulo_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_2X_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_modulo_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_2X_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_2X_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_INC_2X_F64:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_2x_F64(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_INC_2X_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_2x_multiple_times_F64(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_2X_F64:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_2x_F64(res_ptr);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_2X_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_2x_multiple_times_F64(res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_MAC_2X_F64_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_2x_F64(op1_ptr, op2_ptr, op3_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_2X_F64_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_accumulate_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_F64_F64_U32_U32_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_column_major_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, op5, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_MUL_MM_CM_F64_F64_U32_U32_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_matrix_multiple_times_column_major_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, op5, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef op5
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_F64_F64_U32_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_column_major_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MUL_MV_CM_F64_F64_U32_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_matrix_by_vector_multiple_times_column_major_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_F64_F64_U32_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_column_major_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MUL_VM_CM_F64_F64_U32_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_vector_by_matrix_multiple_times_column_major_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_F64_U32_U32_F64:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_F64(op1_ptr, op1_count, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_MTRAN_F64_U32_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_transpose_matrix_multiple_times_F64(op1_ptr, op1_count, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_MINV_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MINV_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_invert_matrix_multiple_times_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MDET_F64_U32_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_F64(op1_ptr, op1_count, op2, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MDET_F64_U32_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_F64(op1_ptr, op1_count, op2, res_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_MDET_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_determinant_multiple_times_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_DOT_F64_F64_U32_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DOT_F64_F64_U32_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DOT_F64_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_dot_product_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_LEN_F64_U32_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_F64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_LEN_F64_U32_F64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((float64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_length_F64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_LEN_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_length_multiple_times_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_DIS_F64_F64_U32_F64_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DIS_F64_F64_U32_F64_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ELE * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((float64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_distance_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_DIS_F64_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_distance_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_NORM_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_NORM_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_normalize_multiple_times_F64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_FORE_F64_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_FORE_F64_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_face_forward_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFL_F64_F64_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFL_F64_F64_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reflect_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_REFR_F64_F64_U32_U32_F64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_REFR_F64_F64_U32_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_refract_multiple_times_F64(op1_ptr, op1_count, op2_ptr, op2_count, op3, op4, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_CABS_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_abs_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CABS_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_abs_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CARG_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_argument_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CARG_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_argument_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CMUL_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_multiply_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CMUL_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_multiply_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CDIV_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_divide_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CDIV_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_divide_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CEXP_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_exp_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CEXP_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_exp_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CLOG_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_log_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CLOG_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_log_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSQRT_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_square_root_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSQRT_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_square_root_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CPOW_F64_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_pow_2x_F64(op1_ptr, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CPOW_F64_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_pow_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CSIN_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_sin_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSIN_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_sin_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOS_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_cos_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOS_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_cos_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTAN_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_tan_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTAN_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_tan_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSINH_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_sinh_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CSINH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_sinh_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOSH_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_cosh_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CCOSH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_cosh_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTANH_F64_F64:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_complex_tanh_2x_F64(op1_ptr, res_ptr);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_CTANH_F64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_complex_tanh_2x_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_IF_EQ_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_EQ_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S08_S08_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S08_S08_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S08_S08_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S08_S08_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_9_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_9_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_17_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_17_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_33_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_33_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_65_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_65_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_129_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_129_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_257_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_257_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_513_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_513_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I08_I08_SCA:
#define INSTR		((qb_instruction_branch_table_1025_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I08_I08_ELE:
#define INSTR		((qb_instruction_branch_table_1025_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_INC_I08_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int8_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I08(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I08_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I08(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I08_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_increment_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_I08_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int8_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I08(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I08_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I08(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I08_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_decrement_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_CBOOL_I08_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_CBOOL_I08_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_ELE_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOV_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_multiple_times_I08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_I08_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I08_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_add_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I08_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_I08_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I08_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I08_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_S08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_S08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BW_AND_I08_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I08_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I08_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_and_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_OR_I08_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I08_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I08_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_or_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_XOR_I08_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I08_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I08_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_xor_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_NOT_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I08_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_not_multiple_times_I08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_EQ_I08_I08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I08_I08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I08_I08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_multiple_times_I08(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_NE_I08_I08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I08_I08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I08_I08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_multiple_times_I08(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_S08_S08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S08_S08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S08_S08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_S08(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_S08_S08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S08_S08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S08_S08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_S08(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_EQ_SET_I08_I08_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_equal_I08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NE_SET_I08_I08_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_not_equal_I08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_SET_S08_S08_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_S08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_S08_S08_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_S08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ABS_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_abs_multiple_times_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_S08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_S08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_S08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_S08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_S08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_S08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S08_S08_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_S08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_S08:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_S08(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_S08:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_S08(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_S08_S08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_S08_S08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S08_S08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S08_S08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S08_S08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S08_S08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S08_S08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S08_S08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_S08_S08_S08_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S08(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_S08_S08_S08_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((int8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S08(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_S08_S08_S08_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S08(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_S08_S08_S08_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S08(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_AFIND_I08_I08_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I08(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I08_I08_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I08(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I08_I08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_multiple_times_I08(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_AFILL_U32_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I08(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I08_I08_ELE:
#define INSTR		((qb_instruction_SCA_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I08(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I08_I08_MIO:
#define INSTR		((qb_instruction_SCA_ARR_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_multiple_times_I08(op1, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_APOS_I08_I08_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_APOS_I08_I08_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I08_I08_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I08_I08_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AREV_I08_U32_I08:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_reverse_I08(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AUNIQ_I08_U32_I08:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_I08(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_SZ_AUNIQ_I08_U32_U32:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_count_I08(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_ACOL_I08_U32_U32_U32_I08:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_array_column_I08(cxt, op1_ptr, op1_count, op2, op3, op4, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ADIFF_I08_I08_U32_I08:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_ADIFF_I08_I08_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_count_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AISECT_I08_I08_U32_I08:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_AISECT_I08_I08_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_count_I08(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_SHUFFLE_U32_I08:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_shuffle_I08(cxt, op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I08_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int8_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I08(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I08_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int8_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I08(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_AREPLACE_I08_S32_S32_U32_U32_U32_I08_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int8_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I08(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_AREPLACE_I08_S32_S32_U32_U32_U32_I08_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int8_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I08(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_ARESIZE1_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int8_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_1_I08(cxt, op1, &op2, &op3, op4, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE2_U32_U32_U32_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	INSTR->operand7
#define res_ptr		(((int8_t *) INSTR->operand8.data_pointer) + INSTR->operand8.index_pointer[0])
#define res_count	INSTR->operand8.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_2_I08(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE3_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	INSTR->operand10
#define res_ptr		(((int8_t *) INSTR->operand11.data_pointer) + INSTR->operand11.index_pointer[0])
#define res_count	INSTR->operand11.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_3_I08(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE4_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	INSTR->operand13
#define res_ptr		(((int8_t *) INSTR->operand14.data_pointer) + INSTR->operand14.index_pointer[0])
#define res_count	INSTR->operand14.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_4_I08(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE5_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	INSTR->operand16
#define res_ptr		(((int8_t *) INSTR->operand17.data_pointer) + INSTR->operand17.index_pointer[0])
#define res_count	INSTR->operand17.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_5_I08(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE6_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	INSTR->operand19
#define res_ptr		(((int8_t *) INSTR->operand20.data_pointer) + INSTR->operand20.index_pointer[0])
#define res_count	INSTR->operand20.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_6_I08(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE7_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	INSTR->operand22
#define res_ptr		(((int8_t *) INSTR->operand23.data_pointer) + INSTR->operand23.index_pointer[0])
#define res_count	INSTR->operand23.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_7_I08(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE8_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	INSTR->operand25
#define res_ptr		(((int8_t *) INSTR->operand26.data_pointer) + INSTR->operand26.index_pointer[0])
#define res_count	INSTR->operand26.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_8_I08(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, &op23, &op24, op25, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef res_ptr
#undef res_count
			
			case QB_APAD_I08_S32_I08_I08:
#define INSTR		((qb_instruction_ARR_SCA_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_pad_I08(op1_ptr, op1_count, op2, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_IF_LT_U08_U08_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_U08_U08_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U08_U08_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U08_U08_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U08(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_MUL_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_U08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U08(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_U08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_U08_U08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U08_U08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U08_U08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_U08(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_U08_U08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U08_U08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U08_U08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_U08(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_SET_U08_U08_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_U08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_U08_U08_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_U08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_U08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_U08(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_U08(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_U08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_U08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_U08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_U08_U08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U08_U08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U08(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U08_U08_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_U08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_U08:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_U08(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_U08:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_U08(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_U08_U08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_U08_U08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U08_U08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U08_U08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U08_U08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U08_U08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U08_U08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U08_U08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_U08_U08_S08_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U08(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_U08_U08_S08_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U08(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_U08_U08_S08_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U08(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_U08_U08_S08_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int8_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U08(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_IF_EQ_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_EQ_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S16_S16_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S16_S16_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S16_S16_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S16_S16_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_9_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_9_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_17_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_17_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_33_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_33_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_65_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_65_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_129_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_129_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_257_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_257_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_513_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_513_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I16_I16_SCA:
#define INSTR		((qb_instruction_branch_table_1025_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I16_I16_ELE:
#define INSTR		((qb_instruction_branch_table_1025_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_INC_I16_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int16_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I16(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I16_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I16(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I16_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_increment_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_I16_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int16_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I16(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I16_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I16(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I16_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_decrement_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_CBOOL_I16_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_CBOOL_I16_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_ELE_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOV_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_multiple_times_I16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_I16_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I16_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_add_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I16_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_I16_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I16_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I16_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_S16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_S16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BW_AND_I16_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I16_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I16_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_and_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_OR_I16_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I16_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I16_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_or_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_XOR_I16_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I16_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I16_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_xor_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_NOT_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I16_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_not_multiple_times_I16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_EQ_I16_I16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I16_I16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I16_I16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_multiple_times_I16(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_NE_I16_I16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I16_I16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I16_I16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_multiple_times_I16(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_S16_S16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S16_S16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S16_S16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_S16(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_S16_S16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S16_S16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S16_S16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_S16(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_EQ_SET_I16_I16_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_equal_I16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NE_SET_I16_I16_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_not_equal_I16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_SET_S16_S16_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_S16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_S16_S16_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_S16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ABS_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_abs_multiple_times_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_S16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_S16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_S16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_S16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_S16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_S16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S16_S16_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_S16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_S16:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_S16(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_S16:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_S16(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_S16_S16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_S16_S16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S16_S16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S16_S16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S16_S16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S16_S16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S16_S16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S16_S16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_S16_S16_S16_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S16(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_S16_S16_S16_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((int16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S16(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_S16_S16_S16_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S16(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_S16_S16_S16_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S16(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_AFIND_I16_I16_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I16(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I16_I16_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I16(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I16_I16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_multiple_times_I16(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_AFILL_U32_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I16(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I16_I16_ELE:
#define INSTR		((qb_instruction_SCA_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I16(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I16_I16_MIO:
#define INSTR		((qb_instruction_SCA_ARR_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_multiple_times_I16(op1, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_APOS_I16_I16_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_APOS_I16_I16_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I16_I16_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I16_I16_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AREV_I16_U32_I16:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_reverse_I16(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AUNIQ_I16_U32_I16:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_I16(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_SZ_AUNIQ_I16_U32_U32:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_count_I16(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_ACOL_I16_U32_U32_U32_I16:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_array_column_I16(cxt, op1_ptr, op1_count, op2, op3, op4, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ADIFF_I16_I16_U32_I16:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_ADIFF_I16_I16_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_count_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AISECT_I16_I16_U32_I16:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_AISECT_I16_I16_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_count_I16(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_SHUFFLE_U32_I16:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_shuffle_I16(cxt, op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I16_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int16_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I16(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I16_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int16_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I16(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_AREPLACE_I16_S32_S32_U32_U32_U32_I16_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int16_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I16(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_AREPLACE_I16_S32_S32_U32_U32_U32_I16_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int16_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I16(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_ARESIZE1_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int16_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_1_I16(cxt, op1, &op2, &op3, op4, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE2_U32_U32_U32_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	INSTR->operand7
#define res_ptr		(((int16_t *) INSTR->operand8.data_pointer) + INSTR->operand8.index_pointer[0])
#define res_count	INSTR->operand8.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_2_I16(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE3_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	INSTR->operand10
#define res_ptr		(((int16_t *) INSTR->operand11.data_pointer) + INSTR->operand11.index_pointer[0])
#define res_count	INSTR->operand11.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_3_I16(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE4_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	INSTR->operand13
#define res_ptr		(((int16_t *) INSTR->operand14.data_pointer) + INSTR->operand14.index_pointer[0])
#define res_count	INSTR->operand14.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_4_I16(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE5_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	INSTR->operand16
#define res_ptr		(((int16_t *) INSTR->operand17.data_pointer) + INSTR->operand17.index_pointer[0])
#define res_count	INSTR->operand17.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_5_I16(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE6_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	INSTR->operand19
#define res_ptr		(((int16_t *) INSTR->operand20.data_pointer) + INSTR->operand20.index_pointer[0])
#define res_count	INSTR->operand20.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_6_I16(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE7_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	INSTR->operand22
#define res_ptr		(((int16_t *) INSTR->operand23.data_pointer) + INSTR->operand23.index_pointer[0])
#define res_count	INSTR->operand23.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_7_I16(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE8_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	INSTR->operand25
#define res_ptr		(((int16_t *) INSTR->operand26.data_pointer) + INSTR->operand26.index_pointer[0])
#define res_count	INSTR->operand26.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_8_I16(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, &op23, &op24, op25, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef res_ptr
#undef res_count
			
			case QB_APAD_I16_S32_I16_I16:
#define INSTR		((qb_instruction_ARR_SCA_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_pad_I16(op1_ptr, op1_count, op2, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_IF_LT_U16_U16_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_U16_U16_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U16_U16_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U16_U16_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U16(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_MUL_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_U16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_U16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U16(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_U16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_U16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_U16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_U16_U16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U16_U16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U16_U16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_U16(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_U16_U16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U16_U16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U16_U16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_U16(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_SET_U16_U16_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_U16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_U16_U16_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_U16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_U16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_U16(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_U16(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_U16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_U16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_U16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_U16_U16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U16_U16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U16(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U16_U16_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_U16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_U16:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_U16(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_U16:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_U16(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_U16_U16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_U16_U16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U16_U16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U16_U16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U16_U16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U16_U16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U16_U16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U16_U16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_U16_U16_S16_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((uint16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U16(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_U16_U16_S16_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((uint16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U16(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_U16_U16_S16_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U16(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_U16_U16_S16_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int16_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U16(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_IF_EQ_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_EQ_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				condition = (op1 == op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_NE_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_not_equal_I64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S64_S64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_S64_S64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_S64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S64_S64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_S64_S64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_S64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_9_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH8_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_9_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 8)) {
					offset = 8;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_17_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH16_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_17_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 16)) {
					offset = 16;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_33_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH32_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_33_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 32)) {
					offset = 32;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_65_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH64_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_65_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 64)) {
					offset = 64;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_129_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH128_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_129_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 128)) {
					offset = 128;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_257_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH256_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_257_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 256)) {
					offset = 256;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_513_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH512_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_513_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 512)) {
					offset = 512;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I64_I64_SCA:
#define INSTR		((qb_instruction_branch_table_1025_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_SWITCH1024_I64_I64_ELE:
#define INSTR		((qb_instruction_branch_table_1025_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				unsigned int offset;
				offset = (unsigned int) (op1 - op2);
				if(UNEXPECTED(offset >= 1024)) {
					offset = 1024;
				}
				handler = INSTR->branch_table[offset].next_handler;
				ip = INSTR->branch_table[offset].instruction_pointer;
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_INC_I64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I64(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_increment_I64(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_INC_I64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_increment_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_DEC_I64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I64(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_decrement_I64(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_DEC_I64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define res_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define res_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_decrement_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef res_ptr
#undef res_count
			
			case QB_CBOOL_I64_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_CBOOL_I64_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				{
					if(!op2) {
						res = FALSE;
					} else {
						res = (op1 != 0);
					}
				}
				ip += sizeof(qb_instruction_ELE_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MOV_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_copy_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_copy_multiple_times_I64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_ADD_I64_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_add_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I64_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_add_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_ADD_I64_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_add_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MUL_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_S64_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[0]
#define res	((int64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_S64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_S64_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((int64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_S64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_S64_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_multiple_times_S64(op1_ptr, op1_count, op2_ptr, op2_count, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_SUB_I64_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I64_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_subtract_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SUB_I64_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_subtract_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_S64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_S64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_S64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_S64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NEG_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_negate_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_NEG_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_negate_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BW_AND_I64_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I64_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_and_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_AND_I64_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_and_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_OR_I64_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I64_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_or_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_OR_I64_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_or_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_XOR_I64_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I64_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_xor_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_BW_XOR_I64_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_xor_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_BW_NOT_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_bitwise_not_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BW_NOT_I64_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_bitwise_not_multiple_times_I64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_EQ_I64_I64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I64_I64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_EQ_I64_I64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_equal_multiple_times_I64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_NE_I64_I64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I64_I64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_I64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_NE_I64_I64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_not_equal_multiple_times_I64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_S64_S64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S64_S64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_S64_S64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_S64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_S64_S64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S64_S64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_S64_S64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_S64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_EQ_SET_I64_I64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_equal_I64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_NE_SET_I64_I64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_not_equal_I64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_SET_S64_S64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_S64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_S64_S64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_S64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_ABS_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_abs_S64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_ABS_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_abs_multiple_times_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_S64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_S64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_S64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_S64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_S64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_S64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_S64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_S64_S64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_S64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_S64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_S64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_S64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_S64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_S64_S64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_S64_S64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S64_S64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_S64_S64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S64_S64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_S64_S64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S64_S64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_S64_S64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_S64_S64_S64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_S64_S64_S64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_S64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_S64_S64_S64_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_S64_S64_S64_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_S64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_AFIND_I64_I64_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I64_I64_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_I64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_AFIND_I64_I64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_search_multiple_times_I64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_AFILL_U32_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I64(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I64_I64_ELE:
#define INSTR		((qb_instruction_SCA_ELE_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_I64(op1, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AFILL_U32_I64_I64_MIO:
#define INSTR		((qb_instruction_SCA_ARR_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_fill_multiple_times_I64(op1, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_APOS_I64_I64_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_APOS_I64_I64_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I64_I64_I32_I32_SCA:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_ARPOS_I64_I64_I32_I32_ELE:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res	((int32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_subarray_position_from_end_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AREV_I64_U32_I64:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_reverse_I64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_AUNIQ_I64_U32_I64:
#define INSTR		((qb_instruction_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_I64(op1_ptr, op1_count, op2, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
			
			case QB_SZ_AUNIQ_I64_U32_U32:
#define INSTR		((qb_instruction_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_unique_count_I64(op1_ptr, op1_count, op2, &res);
				ip += sizeof(qb_instruction_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res
			
			case QB_ACOL_I64_U32_U32_U32_I64:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_array_column_I64(cxt, op1_ptr, op1_count, op2, op3, op4, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ADIFF_I64_I64_U32_I64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_ADIFF_I64_I64_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_difference_count_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_AISECT_I64_I64_U32_I64:
#define INSTR		((qb_instruction_ARR_ARR_SCA_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_AISECT_I64_I64_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR_SCA_SCA * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_intersect_count_I64(op1_ptr, op1_count, op2_ptr, op2_count, op3, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res
			
			case QB_SHUFFLE_U32_I64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_shuffle_I64(cxt, op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I64_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I64(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_ASLICE_S32_S32_U32_U32_I64_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5_ptr		(((int64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define op5_count	INSTR->operand5.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand6.data_pointer) + INSTR->operand6.index_pointer[0])
#define res_count	INSTR->operand6.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_slice_I64(op1, op2, op3, op4, op5_ptr, op5_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_SCA_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5_ptr
#undef op5_count
#undef res_ptr
#undef res_count
			
			case QB_AREPLACE_I64_S32_S32_U32_U32_U32_I64_SCA:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int64_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I64(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_AREPLACE_I64_S32_S32_U32_U32_U32_I64_ELE:
#define INSTR		((qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	INSTR->operand6
#define res_ptr		(((int64_t *) INSTR->operand7.data_pointer) + INSTR->operand7.index_pointer[0])
#define res_count	INSTR->operand7.count_pointer[0]
#define res_count_ptr	INSTR->operand7.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_array_replace_I64(cxt, op1_ptr, op1_count, op2, op3, &op4, op5, op6, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_ARESIZE1_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_1_I64(cxt, op1, &op2, &op3, op4, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE2_U32_U32_U32_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	INSTR->operand7
#define res_ptr		(((int64_t *) INSTR->operand8.data_pointer) + INSTR->operand8.index_pointer[0])
#define res_count	INSTR->operand8.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_2_I64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE3_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	INSTR->operand10
#define res_ptr		(((int64_t *) INSTR->operand11.data_pointer) + INSTR->operand11.index_pointer[0])
#define res_count	INSTR->operand11.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_3_I64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE4_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	INSTR->operand13
#define res_ptr		(((int64_t *) INSTR->operand14.data_pointer) + INSTR->operand14.index_pointer[0])
#define res_count	INSTR->operand14.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_4_I64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE5_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	INSTR->operand16
#define res_ptr		(((int64_t *) INSTR->operand17.data_pointer) + INSTR->operand17.index_pointer[0])
#define res_count	INSTR->operand17.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_5_I64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE6_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	INSTR->operand19
#define res_ptr		(((int64_t *) INSTR->operand20.data_pointer) + INSTR->operand20.index_pointer[0])
#define res_count	INSTR->operand20.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_6_I64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE7_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	INSTR->operand22
#define res_ptr		(((int64_t *) INSTR->operand23.data_pointer) + INSTR->operand23.index_pointer[0])
#define res_count	INSTR->operand23.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_7_I64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef res_ptr
#undef res_count
			
			case QB_ARESIZE8_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	((uint32_t *) INSTR->operand4.data_pointer)[0]
#define op5	((uint32_t *) INSTR->operand5.data_pointer)[0]
#define op6	((uint32_t *) INSTR->operand6.data_pointer)[0]
#define op7	((uint32_t *) INSTR->operand7.data_pointer)[0]
#define op8	((uint32_t *) INSTR->operand8.data_pointer)[0]
#define op9	((uint32_t *) INSTR->operand9.data_pointer)[0]
#define op10	((uint32_t *) INSTR->operand10.data_pointer)[0]
#define op11	((uint32_t *) INSTR->operand11.data_pointer)[0]
#define op12	((uint32_t *) INSTR->operand12.data_pointer)[0]
#define op13	((uint32_t *) INSTR->operand13.data_pointer)[0]
#define op14	((uint32_t *) INSTR->operand14.data_pointer)[0]
#define op15	((uint32_t *) INSTR->operand15.data_pointer)[0]
#define op16	((uint32_t *) INSTR->operand16.data_pointer)[0]
#define op17	((uint32_t *) INSTR->operand17.data_pointer)[0]
#define op18	((uint32_t *) INSTR->operand18.data_pointer)[0]
#define op19	((uint32_t *) INSTR->operand19.data_pointer)[0]
#define op20	((uint32_t *) INSTR->operand20.data_pointer)[0]
#define op21	((uint32_t *) INSTR->operand21.data_pointer)[0]
#define op22	((uint32_t *) INSTR->operand22.data_pointer)[0]
#define op23	((uint32_t *) INSTR->operand23.data_pointer)[0]
#define op24	((uint32_t *) INSTR->operand24.data_pointer)[0]
#define op25	INSTR->operand25
#define res_ptr		(((int64_t *) INSTR->operand26.data_pointer) + INSTR->operand26.index_pointer[0])
#define res_count	INSTR->operand26.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_resize_8_I64(cxt, op1, &op2, &op3, op4, &op5, &op6, op7, &op8, &op9, op10, &op11, &op12, op13, &op14, &op15, op16, &op17, &op18, op19, &op20, &op21, op22, &op23, &op24, op25, res_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef op5
#undef op6
#undef op7
#undef op8
#undef op9
#undef op10
#undef op11
#undef op12
#undef op13
#undef op14
#undef op15
#undef op16
#undef op17
#undef op18
#undef op19
#undef op20
#undef op21
#undef op22
#undef op23
#undef op24
#undef op25
#undef res_ptr
#undef res_count
			
			case QB_APAD_I64_S32_I64_I64:
#define INSTR		((qb_instruction_ARR_SCA_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	((int32_t *) INSTR->operand2.data_pointer)[0]
#define op3_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_pad_I64(op1_ptr, op1_count, op2, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_SCA_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_IF_LT_U64_U64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LT_U64_U64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_U64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U64_U64_SCA:
#define INSTR		((qb_instruction_branch_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_IF_LE_U64_U64_ELE:
#define INSTR		((qb_instruction_branch_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				int condition;
				handler = INSTR->next_handler1;
				qb_do_branch_on_less_than_equal_U64(&condition, op1, op2);
				if(condition) {
					ip = INSTR->instruction_pointer1;
				} else {
					handler = INSTR->next_handler2;
					ip = INSTR->instruction_pointer2;
				}
				if(UNEXPECTED(windows_timeout_check_counter++ == 1048576)) {
					windows_timeout_check_counter = 0;
					if(*windows_timed_out_pointer) {
						cxt->exit_type = QB_VM_TIMEOUT;
						return;
					}
				}
				break;
			}
#undef INSTR
#undef op1
#undef op2
			
			case QB_MUL_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MUL_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_multiply_multiple_times_U64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAC_U64_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint64_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_U64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_U64_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand4.data_pointer)[INSTR->operand4.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_U64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MAC_U64_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define op3_count	INSTR->operand3.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_multiply_accumulate_multiple_times_U64(op1_ptr, op1_count, op2_ptr, op2_count, op3_ptr, op3_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3_ptr
#undef op3_count
#undef res_ptr
#undef res_count
			
			case QB_DIV_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_U64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_DIV_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_divide_multiple_times_U64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MOD_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_U64(cxt, op1, op2, &res, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ELE_ELE_ELE_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef res
			
			case QB_MOD_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_modulo_multiple_times_U64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_ARR_ARR_ARR_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHL_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_left_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHL_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_left_multiple_times_U64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SHR_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_shift_right_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_SHR_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_shift_right_multiple_times_U64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LT_U64_U64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U64_U64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LT_U64_U64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_multiple_times_U64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LE_U64_U64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U64_U64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_LE_U64_U64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res	((int32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_less_than_equal_multiple_times_U64(op1_ptr, op1_count, op2_ptr, op2_count, &res);
				ip += sizeof(qb_instruction_ARR_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res
			
			case QB_LT_SET_U64_U64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_U64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_LE_SET_U64_U64_I32:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_set_less_than_equal_U64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MIN_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_min_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MIN_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_min_multiple_times_U64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MAX_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_max_U64(op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MAX_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_max_multiple_times_U64(op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_RAND_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_U64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_U64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_RAND_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_multiple_times_U64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_MT_RAND_U64_U64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U64_U64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_U64(cxt, op1, op2, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res
			
			case QB_MT_RAND_U64_U64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_random_mt_multiple_times_U64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SORT_U32_U64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_sort_U64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_RSORT_U32_U64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_reverse_sort_U64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_AMIN_U64_U64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMIN_U64_U64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_min_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U64_U64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_AMAX_U64_U64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_max_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U64_U64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_APROD_U64_U64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_product_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U64_U64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_ASUM_U64_U64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_array_sum_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_RANGE_U64_U64_S64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((uint64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_RANGE_U64_U64_S64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_ARR * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res_ptr		(((uint64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_U64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_SZ_RANGE_U64_U64_S64_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_SZ_RANGE_U64_U64_S64_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE_ELE_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
#define op3	((int64_t *) INSTR->operand3.data_pointer)[INSTR->operand3.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand4.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_range_count_U64(op1, op2, op3, &res);
				ip += sizeof(qb_instruction_ELE_ELE_ELE_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res
			
			case QB_MOV_U32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U32_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U32_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U32_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U32_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U32_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U32_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U32_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U32_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U32_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U32_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U32_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U32_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U32_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U32_I64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_UTF8_DEC_U08_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_utf8decode_U32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_SZ_UTF8_ENC_U32_U32:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_utf8encode_count_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UTF8_ENC_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_utf8encode_U32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_PRN_U32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_U32(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_U32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_U32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_U32_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_U32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_U32_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U32(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U32_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U32(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U32_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_U32(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_U32_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_U32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CONV_STR_U08_U32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_U32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_S32_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S32_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S32_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S32_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S32_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S32_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S32_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S32_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S32_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S32_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S32_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_I32_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I32_I08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I32_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I32_I08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I32_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_I32_I08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_I32_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I32_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I32_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I32_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I32_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_I32_I16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S32_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S32_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S32_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S32_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S32_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S32_I64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BOOL_I32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BOOL_I32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_GATH_2X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_2x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_3X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_3x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_4X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_4x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_5X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_5x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_6X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_6x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_7X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_7x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_8X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_8x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_2X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_2x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_3X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_3x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_4X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_4x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_5X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_5x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_6X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_6x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_7X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_7x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_8X_U32_I32_I32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_8x_I32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CLR_SCA_I32:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clear_scalar_I32(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_CLR_ARR_RSZ_U32_I32:
#define INSTR		((qb_instruction_U32_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
#define res_count_ptr	INSTR->operand2.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_I32(cxt, op1, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ARR_RSZ_DIM_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((int32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_update_dimension_I32(cxt, &op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_I32(cxt, op1, op2, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_DIM_U32_U32_U32_U32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
#define res_count_ptr	INSTR->operand5.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_update_dimension_I32(cxt, op1, op2, &op3, op4, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CCLR_ARR_I32_I32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_array_I32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_CCLR_ELE_U32_U32_I32_I32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_element_I32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_PRN_S32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_S32_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_S32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_S32_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S32(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S32_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S32(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S32_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_S32(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_S32_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_S32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_PACK_LE_I32_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_I32(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_LE_I32_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_I32(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_I32_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_I32(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_I32_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_I32(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_UNPACK_LE_U08_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_I32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_LE_U08_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_I32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_I32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_I32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_I32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_I32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_S32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_S32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_F32_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_U32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_S32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_S08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_U08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_S16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_U16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_S64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_S64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F32_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F32_U64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F32_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F32_U64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BOOL_F32_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BOOL_F32_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_GATH_2X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_2x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_3X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_3x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_4X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_4x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_5X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_5x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_6X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_6x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_7X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_7x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_8X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_8x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_2X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_2x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_3X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_3x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_4X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_4x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_5X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_5x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_6X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_6x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_7X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_7x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_8X_U32_F32_F32:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_8x_F32(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CLR_SCA_F32:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clear_scalar_F32(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_CLR_ARR_RSZ_U32_F32:
#define INSTR		((qb_instruction_U32_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
#define res_count_ptr	INSTR->operand2.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_F32(cxt, op1, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ARR_RSZ_DIM_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((float32_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_update_dimension_F32(cxt, &op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_F32(cxt, op1, op2, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_DIM_U32_U32_U32_U32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((float32_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
#define res_count_ptr	INSTR->operand5.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_update_dimension_F32(cxt, op1, op2, &op3, op4, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CCLR_ARR_I32_F32:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_array_F32(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_CCLR_ELE_U32_U32_I32_F32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float32_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_element_F32(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_PRN_F32_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_F32_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_F32_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_F32_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_F32(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_F32_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_F32(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_F32_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_F32(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_F32_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_F32(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_F32_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((float32_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_F32(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_PACK_LE_F32_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_F32(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_LE_F32_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_F32(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_F32_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_F32(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_F32_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((float32_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_F32(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_UNPACK_LE_U08_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_LE_U08_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_F32(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_F32_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_F32_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_F32(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_F64_U32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_U32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_S32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_S32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_S08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_S08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_U08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_U08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_S16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_S16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_U16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_U16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_S64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_S64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_S64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_S64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_F64_U64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_F64_U64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_F64_U64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_F64_U64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BOOL_F64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BOOL_F64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_GATH_2X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_2x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_3X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_3x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_4X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_4x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_5X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_5x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_6X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_6x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_7X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_7x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_8X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_8x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_2X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_2x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_3X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_3x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_4X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_4x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_5X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_5x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_6X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_6x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_7X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_7x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_8X_U32_F64_F64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_8x_F64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CLR_SCA_F64:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clear_scalar_F64(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_CLR_ARR_RSZ_U32_F64:
#define INSTR		((qb_instruction_U32_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
#define res_count_ptr	INSTR->operand2.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_F64(cxt, op1, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ARR_RSZ_DIM_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((float64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_update_dimension_F64(cxt, &op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_F64(cxt, op1, op2, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_DIM_U32_U32_U32_U32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((float64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
#define res_count_ptr	INSTR->operand5.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_update_dimension_F64(cxt, op1, op2, &op3, op4, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CCLR_ARR_I32_F64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_array_F64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_CCLR_ELE_U32_U32_I32_F64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((float64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_element_F64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_PRN_F64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_F64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_F64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_F64_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_F64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_F64_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_F64(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_F64_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_F64(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_F64_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_F64(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_F64_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((float64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_F64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_PACK_LE_F64_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_F64(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_LE_F64_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_F64(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_F64_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_F64(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_F64_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_F64(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_UNPACK_LE_U08_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_LE_U08_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_F64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_F64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_F64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_F64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_S08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S08_I32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S08_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S08_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S08_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S08_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S08_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S08_I16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S08_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S08_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S08_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S08_I64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BOOL_I08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BOOL_I08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_GATH_2X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_2x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_3X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_3x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_4X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_4x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_5X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_5x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_6X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_6x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_7X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_7x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_8X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_8x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_2X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_2x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_3X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_3x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_4X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_4x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_5X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_5x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_6X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_6x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_7X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_7x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_8X_U32_I08_I08:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_8x_I08(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CLR_SCA_I08:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int8_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clear_scalar_I08(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_CLR_ARR_RSZ_U32_I08:
#define INSTR		((qb_instruction_U32_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
#define res_count_ptr	INSTR->operand2.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_I08(cxt, op1, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ARR_RSZ_DIM_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((int8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_update_dimension_I08(cxt, &op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define res_ptr		(((int8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_I08(cxt, op1, op2, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_DIM_U32_U32_U32_U32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int8_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
#define res_count_ptr	INSTR->operand5.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_update_dimension_I08(cxt, op1, op2, &op3, op4, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CCLR_ARR_I32_I08:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_array_I08(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_CCLR_ELE_U32_U32_I32_I08:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_element_I08(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_PRN_S08_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S08_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S08_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_S08_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_S08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_S08_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S08(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S08_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((int8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S08(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S08_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_S08(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_S08_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_S08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CONV_STR_U08_S08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_S08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_U08_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U08_I32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U08_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U08_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U08_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U08_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U08_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U08_I16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U08_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U08_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U08_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U08_I64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_PRN_STR_U08:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_string_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_APP_STR_U08_U32_U08:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_string_U08(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_FLUSH:
#define INSTR		((qb_instruction * __restrict) ip)
			{
				handler = INSTR->next_handler;
				qb_redirect_flush(cxt, ip, 0);
				ip += sizeof(qb_instruction);
				break;
			}
#undef INSTR
			
			case QB_PRN_U08_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U08_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U08_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_U08_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_U08(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_U08_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U08(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U08_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((uint8_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U08(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U08_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_U08(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_U08_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_U08(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CONV_STR_U08_U08_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_U08_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U08(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_S16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S16_I32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S16_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S16_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S16_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S16_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_I16_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I16_I08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I16_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I16_I08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I16_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_I16_I08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S16_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S16_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S16_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S16_I64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BOOL_I16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BOOL_I16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_GATH_2X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_2x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_3X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_3x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_4X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_4x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_5X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_5x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_6X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_6x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_7X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_7x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_8X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_8x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_2X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_2x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_3X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_3x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_4X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_4x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_5X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_5x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_6X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_6x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_7X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_7x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_8X_U32_I16_I16:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_8x_I16(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CLR_SCA_I16:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int16_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clear_scalar_I16(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_CLR_ARR_RSZ_U32_I16:
#define INSTR		((qb_instruction_U32_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
#define res_count_ptr	INSTR->operand2.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_I16(cxt, op1, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ARR_RSZ_DIM_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((int16_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_update_dimension_I16(cxt, &op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define res_ptr		(((int16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_I16(cxt, op1, op2, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_DIM_U32_U32_U32_U32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int16_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
#define res_count_ptr	INSTR->operand5.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_update_dimension_I16(cxt, op1, op2, &op3, op4, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CCLR_ARR_I32_I16:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_array_I16(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_CCLR_ELE_U32_U32_I32_I16:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int16_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_element_I16(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_PRN_S16_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S16_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S16_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_S16_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_S16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_S16_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S16(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S16_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S16(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S16_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_S16(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_S16_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_S16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_PACK_LE_I16_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_I16(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_LE_I16_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_I16(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_I16_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_I16(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_I16_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((int16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_I16(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_UNPACK_LE_U08_I16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_I16(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_LE_U08_I16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_I16(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_I16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_I16(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_I16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_I16(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_S16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_S16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_U16_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U16_I32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U16_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U16_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U16_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U16_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U16_I64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_I64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U16_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U16_I64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U16_I64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_SZ_UTF8_DEC_U08_U32:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_utf8decode_count_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UTF8_DEC_U08_U16:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_utf8decode_U16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_SZ_UTF8_ENC_U16_U32:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_utf8encode_count_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UTF8_ENC_U16_U08:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_utf8encode_U16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_PRN_U16_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_U16(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U16_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_U16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U16_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_U16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_U16_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_U16(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_U16_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U16(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U16_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((uint16_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U16(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U16_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_U16(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_U16_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint16_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_U16(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CONV_STR_U08_U16_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_U16_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U16(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_I64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I64_I32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I64_I32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_I64_I32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S64_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S64_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S64_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S64_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S64_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S64_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_S64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S64_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_S64_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_S64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_S64_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_I64_I08_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int8_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I64_I08(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I08_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int8_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I64_I08(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I08_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_I64_I08(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_I64_I16_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int16_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I64_I16(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I16_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int16_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_I64_I16(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_I64_I16_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((int16_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_I64_I16(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_BOOL_I64_I32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res	((int32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_BOOL_I64_I32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((int32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_boolean_cast_I64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_GATH_2X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_2x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_3X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_3x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_4X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_4x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_5X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_5x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_6X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_6x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_7X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_7x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_GATH_8X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_gather_8x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_2X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_2x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_3X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_3x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_4X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_4x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_5X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_5x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_6X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_6x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_7X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_7x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_SCAT_8X_U32_I64_I64:
#define INSTR		((qb_instruction_U32_ARR_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define op2_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_scatter_8x_I64(op1, op2_ptr, res_ptr);
				ip += sizeof(qb_instruction_U32_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2_ptr
#undef op2_count
#undef res_ptr
#undef res_count
			
			case QB_CLR_SCA_I64:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((int64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_clear_scalar_I64(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_CLR_ARR_RSZ_U32_I64:
#define INSTR		((qb_instruction_U32_ARR * __restrict) ip)
#define op1	INSTR->operand1
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
#define res_count_ptr	INSTR->operand2.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_I64(cxt, op1, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ARR_RSZ_DIM_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((int64_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_array_resize_update_dimension_I64(cxt, &op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	INSTR->operand3
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_I64(cxt, op1, op2, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CLR_ELE_RSZ_DIM_U32_U32_U32_U32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_U32_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
#define op4	INSTR->operand4
#define res_ptr		(((int64_t *) INSTR->operand5.data_pointer) + INSTR->operand5.index_pointer[0])
#define res_count	INSTR->operand5.count_pointer[0]
#define res_count_ptr	INSTR->operand5.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_clear_element_resize_update_dimension_I64(cxt, op1, op2, &op3, op4, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef op4
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CCLR_ARR_I32_I64:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int32_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_array_I64(op1, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_CCLR_ELE_U32_U32_I32_I64:
#define INSTR		((qb_instruction_SCA_SCA_SCA_ARR * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((int32_t *) INSTR->operand3.data_pointer)[0]
#define res_ptr		(((int64_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_predicate_clear_element_I64(op1, op2, op3, res_ptr, res_count);
				ip += sizeof(qb_instruction_SCA_SCA_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef op3
#undef res_ptr
#undef res_count
			
			case QB_PRN_S64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_S64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_S64_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_S64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_S64_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S64(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S64_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_S64(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_S64_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_S64(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_S64_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((int64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_S64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_PACK_LE_I64_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_I64(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_LE_I64_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_little_endian_I64(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_I64_U08_SCA:
#define INSTR		((qb_instruction_SCA_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[0]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_I64(op1, res_ptr);
				ip += sizeof(qb_instruction_SCA_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_PACK_BE_I64_U08_ELE:
#define INSTR		((qb_instruction_ELE_ARR * __restrict) ip)
#define op1	((int64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res_ptr		(((uint8_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_pack_big_endian_I64(op1, res_ptr);
				ip += sizeof(qb_instruction_ELE_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef res_ptr
#undef res_count
			
			case QB_UNPACK_LE_U08_I64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_I64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_LE_U08_I64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_little_endian_I64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_I64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_I64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_UNPACK_BE_U08_I64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_unpack_big_endian_I64(op1_ptr, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_S64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_S64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((int64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_S64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MOV_U64_F32_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U64_F32(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U64_F32_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float32_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U64_F32(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U64_F32_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U64_F32(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_MOV_U64_F64_SCA:
#define INSTR		((qb_instruction_SCA_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define res	((float64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U64_F64(op1, &res);
				ip += sizeof(qb_instruction_SCA_SCA);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U64_F64_ELE:
#define INSTR		((qb_instruction_ELE_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define res	((float64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_cast_U64_F64(op1, &res);
				ip += sizeof(qb_instruction_ELE_ELE);
				break;
			}
#undef INSTR
#undef op1
#undef res
			
			case QB_MOV_U64_F64_MIO:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res_ptr		(((float64_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define res_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_do_cast_multiple_times_U64_F64(op1_ptr, op1_count, res_ptr, res_count);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res_ptr
#undef res_count
			
			case QB_PRN_U64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_U64(cxt, ip, 0);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_array_element_U64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef op1
			
			case QB_PRN_U64_MIO:
#define INSTR		((qb_instruction_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_variable_multiple_times_U64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
			
			case QB_PRN_DIM_U64_U32:
#define INSTR		((qb_instruction_ARR_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
			{
				handler = INSTR->next_handler;
				qb_redirect_print_multidimensional_variable_U64(cxt, ip, 0);
				ip += sizeof(qb_instruction_ARR_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
			
			case QB_APP_VAR_U64_U32_U08_SCA:
#define INSTR		((qb_instruction_SCA_U32_ARR * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U64(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_SCA_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U64_U32_U08_ELE:
#define INSTR		((qb_instruction_ELE_U32_ARR * __restrict) ip)
#define op1	((uint64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_U64(cxt, op1, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ELE_U32_ARR);
				break;
			}
#undef INSTR
#undef op1
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_U64_U32_U08_MIO:
#define INSTR		((qb_instruction_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2	INSTR->operand2
#define res_ptr		(((uint8_t *) INSTR->operand3.data_pointer) + INSTR->operand3.index_pointer[0])
#define res_count	INSTR->operand3.count_pointer[0]
#define res_count_ptr	INSTR->operand3.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_variable_multiple_times_U64(cxt, op1_ptr, op1_count, op2, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_APP_VAR_DIM_U64_U32_U32_U08:
#define INSTR		((qb_instruction_ARR_ARR_U32_ARR * __restrict) ip)
#define op1_ptr		(((uint64_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define op2_ptr		(((uint32_t *) INSTR->operand2.data_pointer) + INSTR->operand2.index_pointer[0])
#define op2_count	INSTR->operand2.count_pointer[0]
#define op3	INSTR->operand3
#define res_ptr		(((uint8_t *) INSTR->operand4.data_pointer) + INSTR->operand4.index_pointer[0])
#define res_count	INSTR->operand4.count_pointer[0]
#define res_count_ptr	INSTR->operand4.count_pointer
			{
				handler = INSTR->next_handler;
				qb_do_append_multidimensional_variable_U64(cxt, op1_ptr, op1_count, op2_ptr, op2_count, op3, res_ptr, res_count_ptr);
				ip += sizeof(qb_instruction_ARR_ARR_U32_ARR);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef op2_ptr
#undef op2_count
#undef op3
#undef res_ptr
#undef res_count
#undef res_count_ptr
			
			case QB_CONV_STR_U08_U64_SCA:
#define INSTR		((qb_instruction_ARR_SCA * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_SCA);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_CONV_STR_U08_U64_ELE:
#define INSTR		((qb_instruction_ARR_ELE * __restrict) ip)
#define op1_ptr		(((uint8_t *) INSTR->operand1.data_pointer) + INSTR->operand1.index_pointer[0])
#define op1_count	INSTR->operand1.count_pointer[0]
#define res	((uint64_t *) INSTR->operand2.data_pointer)[INSTR->operand2.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_convert_from_string_U64(op1_ptr, op1_count, &res);
				ip += sizeof(qb_instruction_ARR_ELE);
				break;
			}
#undef INSTR
#undef op1_ptr
#undef op1_count
#undef res
			
			case QB_MUL_MM_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate_matrix_by_matrix_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_MUL_MV_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate_matrix_by_vector_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_MUL_VM_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate_vector_by_matrix_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_SM_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate_square_matrix_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_TRAN_GUARD_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate_transform_U32(cxt, op1, op2, op3, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
			
			case QB_CROSS_2X_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate2dcross_product_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_CROSS_3X_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate3dcross_product_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_CROSS_4X_GUARD_U32_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
#define op3	((uint32_t *) INSTR->operand3.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate4dcross_product_U32(cxt, op1, op2, op3, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
#undef op3
			
			case QB_VV_GUARD_U32_U32:
#define INSTR		((qb_instruction_SCA_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
#define op2	((uint32_t *) INSTR->operand2.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				if(!qb_do_validate_vector_widths_U32(cxt, op1, op2, line_id)) {
					return;
				}
				ip += sizeof(qb_instruction_SCA_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
#undef op2
			
			case QB_TIME_F64_SCA:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_do_get_time_F64(&res);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef res
			
			case QB_TIME_F64_ELE:
#define INSTR		((qb_instruction_ELE * __restrict) ip)
#define res	((float64_t *) INSTR->operand1.data_pointer)[INSTR->operand1.index_pointer[0]]
			{
				handler = INSTR->next_handler;
				qb_do_get_time_F64(&res);
				ip += sizeof(qb_instruction_ELE);
				break;
			}
#undef INSTR
#undef res
			
			case QB_EXT_U32:
#define INSTR		((qb_instruction_SCA_line_id * __restrict) ip)
#define line_id		INSTR->line_id
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_run_zend_extension_op(cxt, op1, line_id);
				ip += sizeof(qb_instruction_SCA_line_id);
				break;
			}
#undef INSTR
#undef line_id
#undef op1
			
			case QB_DBG_SYNC_U32:
#define INSTR		((qb_instruction_SCA * __restrict) ip)
#define op1	((uint32_t *) INSTR->operand1.data_pointer)[0]
			{
				handler = INSTR->next_handler;
				qb_sync_shadow_variable(cxt, op1);
				ip += sizeof(qb_instruction_SCA);
				break;
			}
#undef INSTR
#undef op1
			
			default:
			__assume(0);
		}
	} while(1);
}

#if NATIVE_COMPILE_ENABLED && ZEND_DEBUG
#include "qb_native_proc_debug.c"
#ifdef HAVE_NATIVE_PROC_RECORDS
qb_native_proc_record *native_proc_table = native_proc_records;
uint32_t native_proc_table_size = sizeof(native_proc_records) / sizeof(qb_native_proc_record);
#else
qb_native_proc_record *native_proc_table = NULL;
uint32_t native_proc_table_size = 0;
#endif
#endif
